"""Module contains code pertaining to CruxClient."""

from typing import (  # noqa: F401 pylint: disable=unused-import
    Any,
    Dict,
    List,
    MutableMapping,
    Optional,
    Text,
    Tuple,
    Union,
)

from requests.exceptions import (
    ConnectTimeout,
    HTTPError,
    ProxyError,
    ReadTimeout,
    SSLError,
    TooManyRedirects,
)

from crux._config import CruxConfig
from crux._utils import create_logger, Headers, url_builder
from crux.exceptions import (
    CruxAPIError,
    CruxClientConnectionError,
    CruxClientHTTPError,
    CruxClientTimeout,
    CruxClientTooManyRedirects,
    CruxResourceNotFoundError,
)


log = create_logger(__name__)


class CruxClient(object):
    """Crux HTTP REST client."""

    def __init__(self, crux_config):
        # type: (CruxConfig) -> None
        if crux_config is None:
            log.debug("crux_config is None, initializing CruxConfig object")
            self.crux_config = CruxConfig()  # type: CruxConfig
        else:
            log.debug("Using the passed crux_config object")
            self.crux_config = crux_config

    def api_call(  # pylint: disable=too-many-branches, too-many-statements
        self,
        method,  # type: str
        path,  # type: List[str]
        model=None,  # type: Any
        headers=None,  # type: MutableMapping[Text, Text]
        params=None,  # type: Dict[Any,Any]
        json=None,  # type: Dict[Any,Any]
        data=None,  # type: Dict[Any,Any]
        stream=False,  # type: bool
        connect_timeout=9.5,  # type: float
        read_timeout=60,  # type: float
    ):
        # type:(...) -> Any
        """
        Requests and Serializes response from API Backend.

        Args:
            method (str): REST method name.
            path (str): API resource path.
            model (crux.models.CruxModel): Deserialization Model. Defaults to None.
            headers (dict): Additonal header parameters. Defaults to None.
            json (dict): Body data to be passed with request. Defaults to None.
            params (dict): Data to be passed in query string. Defaults to None.
            data (dict): Should be used while passing form encoded data. Defaults to None.
            stream (bool): Should be set to True, when response is required to be streamed.
                Defaults to False.
            connect_timeout (float): Request connect timeout configuration in seconds.
                Defaults to 60.5.
            read_timeout (float): Request read timeout configuration in seconds.
                Defaults to 60.

        Returns:
            crux.models.Model or bool: Serialized response from API backend.

        Raises:
            TypeError: If Path is not of list type.
            CruxClientHTTPError: If there is HTTP related error.
            CruxClientConnectionError: If there is SSL or Proxy related error.
            CruxClientTimeout: If there is timout related error.
            CruxResourceNotFoundError: If API has status code 400.
            CruxAPIError: If API has status code other than 2XX.
        """

        if path is None or not isinstance(path, list):
            raise TypeError("Path cannot be of NoneType. It should be of Type List")

        url = url_builder(
            url_base=self.crux_config.api_host,
            url_prefix=self.crux_config.api_prefix,
            url_path_list=path,
        )

        if headers is None:
            headers = Headers({})

        if params is None:
            params = {}

        auth_scheme = "Bearer"  # type: str
        bearer_token = "{scheme} {key}".format(
            scheme=auth_scheme, key=self.crux_config.api_key
        )  # type: Text

        user_agent = self.crux_config.user_agent  # type: Text

        headers["authorization"] = bearer_token
        headers["user-agent"] = user_agent

        session = self.crux_config.session

        if method in ("GET", "DELETE", "PUT", "POST"):
            try:
                log.trace("Setting request stream: %s", stream)
                log.trace("Setting request data: %s, json: %s", data, json)
                log.trace("Setting request params: %s", params)
                log.trace("Setting headers: %s", headers)
                response = session.request(
                    method,
                    url,
                    headers=headers,
                    data=data,
                    json=json,
                    stream=stream,
                    params=params,
                    timeout=(connect_timeout, read_timeout),
                )
            except HTTPError as err:
                raise CruxClientHTTPError(str(err), err.response)
            except TooManyRedirects as err:
                raise CruxClientTooManyRedirects(str(err))
            except (ProxyError, SSLError) as err:
                raise CruxClientConnectionError(str(err))
            except (ConnectTimeout, ReadTimeout) as err:
                raise CruxClientTimeout(str(err))
        else:
            raise ValueError("Request Method Type should be in GET, DELETE, PUT, POST")

        if response.status_code in (200, 201, 202, 206):
            if model is None:
                log.debug("Model is set to None, returning response dictionary")
                return response
            else:
                if isinstance(response.json(), list):
                    log.debug("Response is list of type %s", model)
                    serial_list = []
                    for item in response.json():
                        obj = model.from_dict(item)
                        obj.connection = self
                        obj.raw_response = response.json()
                        serial_list.append(obj)
                    return serial_list

                else:
                    log.debug("Response is of type %s", model)
                    obj = model.from_dict(response.json())
                    obj.connection = self
                    obj.raw_response = response.json()
                    return obj
        elif response.status_code == 204:
            log.debug("Response code is 204, returning True boolean value")
            return True
        else:
            if response.status_code == 404:
                raise CruxResourceNotFoundError(response.json())
            raise CruxAPIError(response.json())

    def close(self):
        """Closes the Session."""
        self.crux_config.session.close()



"""Module contains Label model."""

from typing import Any, Dict, Optional  # noqa: F401

from crux.models.model import CruxModel


class Label(CruxModel):
    """Label Model."""

    def __init__(self, label_key=None, label_value=None):
        # type: (str, str) -> None
        """
        Attributes:
            label_key (str): Label Key. Defaults to None.
            label_value (str): Label value. Defaults to None.
        """
        self.label_key = label_key
        self.label_value = label_value

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Label object to Label Dictionary.

        Returns:
            dict: Label Dictionary.
        """
        return {"labelKey": self.label_key, "labelValue": self.label_value}

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str,str]) -> Label
        """Transforms Label Dictionary to Label object.

        Args:
            a_dict (dict): Label Dictionary.

        Returns:
            crux.models.Label: Label Object.
        """
        label_key = a_dict["labelKey"]
        label_value = a_dict["labelValue"]

        return cls(label_key=label_key, label_value=label_value)



"""Module contains Resource model."""

from enum import Enum
import os
import posixpath
from typing import Any, Dict, List, Union  # noqa: F401

from requests.models import Response  # noqa: F401 pylint: disable=unused-import

from crux._utils import DEFAULT_CHUNK_SIZE, Headers
from crux.models.model import CruxModel
from crux.models.permission import Permission


class Resource(CruxModel):
    """Resource Model."""

    def __init__(
        self,
        id=None,  # type: str # id name is by design pylint: disable=redefined-builtin
        dataset_id=None,  # type: str
        folder_id=None,  # type: str
        folder=None,  # type: str
        name=None,  # type: str
        size=None,  # type: str
        type=None,  # type: str # type name is by design pylint: disable=redefined-builtin
        config=None,  # type: Dict[str, Any]
        provenance=None,  # type: str
        as_of=None,  # type: str
        created_at=None,  # type: str
        modified_at=None,  # type: str
        storage_id=None,  # type: str
        description=None,  # type: str
        media_type=None,  # type: str
        tags=None,  # type: List[str]
        labels=None,  # type: Dict[str, str]
        connection=None,
        raw_response=None,  # type: Dict[Any, Any]
    ):
        # type: (...) -> None
        """
        Attributes:
            id (str): Resourece Id. Defaults to None.
            dataset_id (str): Dataset Identity. Defaults to None.
            folder_id (str): Folder Identity. Defaults to None.
            folder (str): Folder Name. Defaults to None.
            name (str): Resource name. Defaults to None.
            size (str): Resource size. Defaults to None.
            type (str): Resource type. Defaults to None.
            config (str): Resource config. Defaults to None.
            provenance (list): Resource Provenance. Defaults to None.
            as_of (str): Resource as_of. Defaults to None.
            created_at (str): Resource created_at. Defaults to None.
            modified_at (str): Resource modified_at. Defaults to None.
            storage_id (str): Resource storage Identity. Defaults to None.
            description (str): Resource description. Defaults to None.
            media_type (str): Resource Media Type. Defaults to None.
            tags (:obj:`list` of :obj:`str`): Resource tags. Defaults to None.
            labels (dict): Dictionary containing Label Key and Values.
                Defaults to None.
            connection (crux._client.CruxClient): Connection Object. Defaults to None.
            raw_response (dict): Response Content. Defaults to None.

        Raises:
            ValueError: If name or tags are set to None.
            TypeError: If tags are not of list type.
        """
        self._id = id
        self._dataset_id = dataset_id
        self._folder_id = folder_id
        self._description = description
        self._name = name
        self._size = size if size else 0
        self._type = type
        self._config = config
        self._provenance = provenance
        self._as_of = as_of
        self._created_at = created_at
        self._storage_id = storage_id
        self._media_type = media_type
        self._modified_at = modified_at
        self._tags = tags
        self._folder = None
        self._labels = labels if labels else {}  # type: Dict[str, str]
        self._folder = folder

        self.connection = connection
        self.raw_response = raw_response

    @property
    def id(self):
        """str: Gets the Resource ID."""
        return self._id

    @property
    def description(self):
        """str: Gets the Resource Description."""
        return self._description

    @property
    def media_type(self):
        """str: Gets the Resource Description."""
        return self._media_type

    @property
    def dataset_id(self):
        """str: Gets the Dataset ID."""
        return self._dataset_id

    @property
    def folder_id(self):
        """str: Gets the Folder ID."""
        return self._folder_id

    @property
    def storage_id(self):
        """str: Gets the Storage ID."""
        return self._storage_id

    @property
    def name(self):
        """str: Gets the Resource Name."""
        return self._name

    @property
    def config(self):
        """str: Gets the config."""
        return self._config

    @property
    def provenance(self):
        """str: Gets the Provenance."""
        return self._provenance

    @property
    def type(self):
        """str: Gets the Resource Type."""
        return self._type

    @property
    def tags(self):
        """:obj:`list` of :obj:`str`: Gets the Resource Tags."""
        return self._tags

    @property
    def labels(self):
        """dict: Gets the Resource labels."""
        return self._labels

    @property
    def as_of(self):
        """str: Gets the as_of."""
        return self._as_of

    @property
    def created_at(self):
        """str: Gets created_at."""
        return self._created_at

    @property
    def modified_at(self):
        """str: Gets modified_at."""
        return self._modified_at

    @property
    def size(self):
        """int: Gets the size."""
        return int(self._size)

    @property
    def path(self):
        """str: Compute or Get the resource path."""
        return posixpath.join(self.folder, self.name)

    @property
    def folder(self):
        """str: Compute or Get the folder name."""
        if self._folder:
            return self._folder

        self._folder = self._get_folder()
        return self._folder

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Resource object to Resource Dictionary.

        Returns:
            dict: Resource Dictionary.
        """
        return {
            "resourceId": self.id,
            "datasetId": self.dataset_id,
            "description": self.description,
            "folderId": self.folder_id,
            "mediaType": self.media_type,
            "name": self.name,
            "size": self.size,
            "type": self.type,
            "config": self.config,
            "provenance": self.provenance,
            "asOf": self.as_of,
            "tags": self.tags,
            "labels": self.labels,
            "storageId": self.storage_id,
            "createdAt": self.created_at,
            "modifiedAt": self.modified_at,
        }

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, Any]) -> Any
        """Transforms Resource Dictionary to Resource object.

        Args:
            a_dict (dict): Resource Dictionary.

        Returns:
            crux.models.Resource: Resource Object.
        """
        id = a_dict[  # id name is by design pylint: disable=redefined-builtin
            "resourceId"
        ]
        dataset_id = a_dict["datasetId"]
        description = a_dict["description"]
        folder_id = a_dict["folderId"]
        storage_id = a_dict["storageId"]
        media_type = a_dict["mediaType"]
        description = a_dict["description"]
        name = a_dict["name"]
        if "tags" in a_dict:
            tags = a_dict["tags"]
        else:
            tags = []
        type = a_dict[  # type name is by design pylint: disable=redefined-builtin
            "type"
        ]
        if "config" in a_dict:
            config = a_dict["config"]
        else:
            config = None
        if "labels" in a_dict:
            labels = {}
            for label in a_dict["labels"]:
                labels[label["labelKey"]] = label["labelValue"]
        else:
            labels = {}
        provenance = a_dict["provenance"]
        created_at = a_dict["createdAt"]
        modified_at = a_dict["modifiedAt"]
        size = a_dict["size"]
        as_of = a_dict["asOf"]

        return cls(
            dataset_id=dataset_id,
            id=id,
            folder_id=folder_id,
            media_type=media_type,
            storage_id=storage_id,
            description=description,
            name=name,
            tags=tags,
            labels=labels,
            type=type,
            config=config,
            provenance=provenance,
            created_at=created_at,
            modified_at=modified_at,
            size=size,
            as_of=as_of,
        )

    def delete(self):
        # type: () -> bool
        """Deletes Resource from Dataset.

        Returns:
            bool: True if it is deleted.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "DELETE", ["resources", self.id], headers=headers
        )

    def update(self, name=None, description=None, tags=None, provenance=None):
        # type: (str, str, List[str], str) -> bool
        """Updates the metadata for Resource.

        Args:
            name (str): Name of resource. Defaults to None.
            description (str): Description of the resource. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags. Defaults to None.
            provenance (str): Provenance for a resource. Defaults to None.

        Returns:
            bool: True, if resource is updated.

        Raises:
            ValueError: It is raised if name, description or tags are unset.
            TypeError: It is raised if tags are not of type List.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        body = {}  # type: Dict[str, Union[str, List, Dict]]

        if name is not None:
            body["name"] = name
        if description is not None:
            body["description"] = description
        if tags is not None:
            if isinstance(tags, list):
                body["tags"] = tags
            else:
                raise TypeError("Tags should be of type list")
        if provenance is not None:
            body["provenance"] = provenance

        if body:
            response = self.connection.api_call(
                "PUT", ["resources", self.id], headers=headers, json=body
            )

            response_dict = response.json()

            if "name" in response_dict:
                self._name = response_dict["name"]
            if "tags" in response_dict and tags is not None:
                self._tags = response_dict["tags"]
            if "description" in response_dict:
                self._description = response_dict["description"]
            if "provenance" in response_dict:
                self._provenance = response_dict["provenance"]
            return True
        else:
            raise ValueError("Name, Description, Tags or Provenance should be set")

    def add_permission(self, identity_id, permission):
        # type: (str, str) -> Union[bool, Permission]
        """Adds permission to the resource.

        Args:
            identity_id: Identity Id to be set.
            permission: Permission to be set.

        Returns:
            crux.models.Permission: Permission Object.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "PUT",
            ["permissions", self.id, identity_id, permission],
            model=Permission,
            headers=headers,
        )

    def delete_permission(self, identity_id, permission):
        # type: (str, str) -> bool
        """Deletes permission from the resource.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "DELETE", ["permissions", self.id, identity_id, permission], headers=headers
        )

    def list_permissions(self):
        # type: () -> List[Permission]
        """Lists the permission on the resource.

        Returns:
            list (:obj:`crux.models.Permission`): List of Permission Objects.
        """
        headers = Headers({"accept": "application/json"})
        return self.connection.api_call(
            "GET",
            ["resources", self.id, "permissions"],
            model=Permission,
            headers=headers,
        )

    def add_label(self, label_key, label_value):
        # type: (str, str) -> bool
        """Adds label to Resource.

        Args:
            label_key (str): Label Key for Resource.
            label_value (str): Label Value for Resource.

        Returns:
            bool: True if label is added, False otherwise.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        response_result = self.connection.api_call(
            "PUT",
            [
                "datasets",
                self.dataset_id,
                "resources",
                self.id,
                "labels",
                label_key,
                label_value,
            ],
            headers=headers,
        )

        if response_result:
            self._labels[label_key] = label_value
            return True
        else:
            return False

    def delete_label(self, label_key):
        # type: (str) -> bool
        """Deletes label from Resource.

        Args:
            label_key (str): Label Key for Resource.

        Returns:
            bool: True if label is deleted, False otherwise.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        response_result = self.connection.api_call(
            "DELETE",
            ["datasets", self.dataset_id, "resources", self.id, "labels", label_key],
            headers=headers,
        )

        if response_result:
            try:
                del self._labels[label_key]
            except KeyError:
                return False
            return True
        else:
            return False

    def add_labels(self, labels_dict):
        # type: (dict) -> bool
        """Adds multiple labels to Resource.

        Args:
            label_dict (dict): Labels (key/value pairs) to add to the Resource.

        Returns:
            bool: True if the labels were added, False otherwise.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        labels_list = []
        for label_key, label_value in labels_dict.items():
            if label_key is not None and label_value is not None:
                label_key = (
                    label_key.value if isinstance(label_key, Enum) else label_key
                )
                labels_list.append(
                    {"labelKey": str(label_key), "labelValue": str(label_value)}
                )

        data = {"labels": labels_list}

        response_result = self.connection.api_call(
            "PUT",
            ["datasets", self.dataset_id, "resources", self.id, "labels"],
            headers=headers,
            json=data,
        )

        if response_result:
            for label_key, label_value in labels_dict.items():
                self._labels[label_key] = label_value
            return True
        else:
            return False

    def _get_folder(self):
        # type: () -> str
        """Fetches the folder of the resource.

        Returns:
            str: Folder name of the resource.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        response = self.connection.api_call(
            "GET", ["resources", self.id, "folderpath"], headers=headers
        )

        return response.json().get("path")

    def _download(self, file_obj, media_type, chunk_size=DEFAULT_CHUNK_SIZE):

        if media_type is not None:
            headers = Headers({"accept": media_type})
        else:
            headers = None

        data = self.connection.api_call(
            "GET", ["resources", self.id, "content"], headers=headers, stream=True
        )

        for chunk in data.iter_content(chunk_size=chunk_size):
            file_obj.write(chunk)
        data.close()
        return True

    def refresh(self):
        """Refresh Resource model from API backend.

        Returns:
            bool: True, if it is able to refresh the model,
                False otherwise.
        """
        # type () -> bool
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        resource_object = self.connection.api_call(
            "GET", ["resources", self.id], headers=headers, model=Resource
        )
        self.__dict__.update(resource_object.__dict__)
        return True


# https://github.com/python/mypy/issues/2477, mypy is performing checking with Python2
class MediaType(Enum):  # type: ignore
    """MediaType Enumeration Model."""

    JSON = "application/json"
    NDJSON = "application/x-ndjson"
    CSV = "text/csv"
    PARQUET = "application/parquet"
    AVRO = "avro/binary"

    @classmethod
    def detect(cls, file_name):
        # type: (str) -> str
        """Detects the media_type from the file extension.

        Args:
            file_name (str): Absolute or Relative Path of the file.

        Returns:
            str: MediaType extension.

        Raises:
            LookupError: If file type is not supported.
        """
        file_ext = os.path.splitext(file_name)[1][1:].upper()

        if file_ext in cls.__members__:
            return cls[file_ext].value  # type: ignore
        else:
            raise LookupError("File/Media Type not supported.")



"""Module contains Table model."""

from typing import Any, Dict  # noqa: F401

from crux._compat import unicode
from crux._utils import create_logger, DEFAULT_CHUNK_SIZE
from crux.models.resource import Resource


log = create_logger(__name__)


class Table(Resource):
    """Table model."""

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Table object to Table Dictionary.

        Returns:
            dict: Table Dictionary.
        """
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "type": self.type,
            "config": self.config,
            "folder": self.folder,
        }

    def download(self, dest, media_type, chunk_size=DEFAULT_CHUNK_SIZE):
        # type: (str, str, int) -> bool
        """Downloads the table resource.

        Args:
            dest (str or file): Local OS path at which file resource will be downloaded.
            media_type (str): Content Type for download.
            chunk_size (int): Number of bytes to be read in memory.

        Returns:
            bool: True if it is downloaded.

        Raises:
            TypeError: If dest is not a file like or string type.
        """
        if hasattr(dest, "write"):
            return self._download(dest, media_type=media_type, chunk_size=chunk_size)
        elif isinstance(dest, (str, unicode)):
            with open(dest, "wb") as file_obj:
                return self._download(
                    file_obj, media_type=media_type, chunk_size=chunk_size
                )
        else:
            raise TypeError("Invalid Data Type for dest: {}".format(type(dest)))



"""Module contains Crux object to interact with root APIs."""

from typing import (  # noqa: F401 pylint: disable=unused-import
    Dict,
    List,
    MutableMapping,
    Optional,
    Text,
    Union,
)

from crux._client import CruxClient
from crux._config import CruxConfig
from crux._utils import Headers
from crux.models import Dataset, File, Folder, Identity, Job, Query, Table
from crux.models._factory import get_resource_object


class Crux(object):
    """Crux APIs."""

    def __init__(
        self,
        api_key=None,  # type: Optional[str]
        api_host=None,  # type: str
        proxies=None,  # type: Optional[MutableMapping[Text, Text]]
        user_agent=None,  # type: str
        api_prefix=None,  # type: str
        only_use_crux_domains=None,  # type: bool
    ):
        # type: (...) -> None
        crux_config = CruxConfig(
            api_key=api_key,
            api_host=api_host,
            proxies=proxies,
            user_agent=user_agent,
            api_prefix=api_prefix,
            only_use_crux_domains=only_use_crux_domains,
        )

        self.api_client = CruxClient(crux_config=crux_config)

    def close(self):
        """Closes the Connection."""
        self.api_client.close()

    def whoami(self):
        # type: () -> Identity
        """Returns the Identity of Current User.

        Returns:
            crux.models.Identity: Identity object.
        """
        headers = Headers(
            {"accept": "application/json"}
        )  # type: Optional[MutableMapping[Text, Text]]
        return self.api_client.api_call(
            "GET", ["identities", "whoami"], model=Identity, headers=headers
        )

    def create_dataset(self, name, description=None, tags=None):
        # type: (str, str, List[str]) -> Dataset
        """Creates the Dataset.

        Args:
            name (str): Sets whether to sort or not.
            description (str): Folder for which resource should be listed. Defaults to None.
            tags (:obj:`list` of :obj:`str`): Sets the offset. Defaults to None.

        Returns:
            crux.models.Dataset: Dataset object.
        """

        tags = tags if tags else []

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )  # type: MutableMapping[Text, Text]
        dataset = Dataset(name=name, description=description, tags=tags)
        return self.api_client.api_call(
            "POST", ["datasets"], json=dataset.to_dict(), model=Dataset, headers=headers
        )

    def get_dataset(self, id):  # id name is by design pylint: disable=redefined-builtin
        # type: (str) -> Dataset
        """Fetches the Dataset.

        Args:
            id (str): Dataset ID which is to be fetched.

        Returns:
            crux.models.Dataset: Dataset object
        """
        headers = Headers(
            {"accept": "application/json"}
        )  # type: MutableMapping[Text, Text]
        return self.api_client.api_call(
            "GET", ["datasets", id], model=Dataset, headers=headers
        )

    def get_resource(self, id):  # id is by design pylint: disable=redefined-builtin
        # type: (str) -> Union[File, Folder, Query, Table]
        """Fetches the Resource by ID.

        Any supported resource can be fetched. The object returned will be an instance
        of the correct subclass, for example a ``crux.models.File`` instance will be
        returned for file resources.

        Args:
            id (str): Resource ID which is to be fetched.

        Returns:
            crux.models.Resource: Resource or its Child Object.
        """
        headers = Headers(
            {"accept": "application/json"}
        )  # type: MutableMapping[Text, Text]

        response = self.api_client.api_call("GET", ["resources", id], headers=headers)
        raw_resource = response.json()

        resource = get_resource_object(
            resource_type=raw_resource.get("type"), data=raw_resource
        )
        resource.connection = self.api_client
        resource.raw_response = raw_resource
        return resource

    def _call_drives_my(self):
        headers = Headers(
            {"accept": "application/json"}
        )  # type: MutableMapping[Text, Text]

        response = self.api_client.api_call(
            "GET", ["drives", "my"], model=None, headers=headers
        )

        return response.json()

    def list_datasets(self, owned=True, subscribed=True):
        # type: (bool, bool) -> List[Dataset]
        """Fetches a list of owned and/or subscribed Datasets.

        Args:
            owned (bool): Show datasets owned by the caller. Defaults to True.
            subscribed (bool): Show datasets the user has a subscription. Defaults to True.

        Returns:
            list(:obj:`crux.models.Dataset`): List of Dataset objects.
        """
        datasets = self._call_drives_my()
        dataset_list = []

        if owned:
            for dataset in datasets["owned"]:
                obj = Dataset.from_dict(dataset)
                dataset_list.append(obj)

        if subscribed:
            for dataset in datasets["subscriptions"]:
                obj = Dataset.from_dict(dataset)
                dataset_list.append(obj)

        return dataset_list

    def list_public_datasets(self):
        # type: () -> List[Dataset]
        """Fetches a list of public Datasets.

        Returns:
            list (:obj:`crux.models.Dataset`): List of Dataset objects.
        """
        headers = Headers(
            {"accept": "application/json"}
        )  # type: MutableMapping[Text, Text]
        return self.api_client.api_call(
            "GET", ["datasets", "public"], model=Dataset, headers=headers
        )

    def get_job(self, job_id):
        # type: (str) -> Job
        """Fetches the Job.

        Args:
            job_id (str): Job ID which is to be fetched.

        Returns:
            crux.models.Job: Job object.
        """
        headers = Headers(
            {"accept": "application/json"}
        )  # type: MutableMapping[Text, Text]
        return self.api_client.api_call(
            "GET", ["jobs", job_id], model=Job, headers=headers
        )

    def set_datasets_provenance(self, provenance):
        # type(Dict[Any, Any]) -> Dict[Any, Any]
        """ Sets the Dataset Provenance

        Args:
            provenance (dict): Provenance dictionary

        Returns:
            dict: Provenance response dictionary.

        Example:
            .. code-block:: python

                from crux import Crux

                conn = Crux()

                provenance = {
                    "dataset_id":[
                        {
                            "workflow_id": "test_id",
                            "pipeline_ids": ["test_id_1","test_id_2"],
                            "cron_spec": "0 0 1 1 0"
                            }
                        ]
                    }
                response = conn.set_datasets_provenance(provenance=provenance)
        """
        headers = Headers({"accept": "application/json"})

        response = self.api_client.api_call(
            "POST", ["datasets", "provenance"], headers=headers, json=provenance
        )
        return response.json()



"""Module contains Query model."""

from typing import Any, Dict, Iterable  # noqa: F401 pylint: disable=unused-import

from crux._utils import DEFAULT_CHUNK_SIZE, Headers, valid_chunk_size
from crux.models.resource import Resource


class Query(Resource):
    """Query Model."""

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Query object to Query dictionary.

        Returns:
            dict: Query dictionary.
        """
        return {
            "name": self.name,
            "description": self.description,
            "tags": self.tags,
            "type": self.type,
            "config": self.config,
            "folder": self.folder,
        }

    def run(
        self,
        format="csv",  # type: str # format name is by design pylint: disable=redefined-builtin
        params=None,  # type: Dict[str, str]
        chunk_size=DEFAULT_CHUNK_SIZE,  # type: int
        decode_unicode=False,  # type: bool
    ):
        # type(...) -> Iterable[str]
        """Method which streams the Query

        Args:
            format (str): Output format of the query. Defaults to csv.
            params (dict): Run parameters. Defaults to None.
            chunk_size (int): Chunk Size for the stream
            decode_unicode (bool): If decode_unicode is True,content will be decoded using the
                best available encoding based on the response.
                Defaults to False.

        Yields:
            bytes: Bytes of content.

        Raises:
            ValueError: If chunk size is not multiple of 256 KiB.
        """

        params = params if params else {}

        headers = Headers({"content-type": "application/json", "accept": "*/*"})

        params["format"] = format

        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        data = self.connection.api_call(
            "GET",
            ["resources", self.id, "content"],
            params=params,
            stream=True,
            headers=headers,
        )

        return data.iter_content(chunk_size=chunk_size, decode_unicode=decode_unicode)

    def download(
        self, dest, format="csv", params=None
    ):  # It is by design pylint: disable=redefined-builtin
        # type: (str, str, Dict[Any, Any]) -> bool
        """Method which streams the Query

        Args:
            dest (str): Local OS path at which resource will be downloaded.
            media_type (str): Output format of the query. Defaults to csv.
            params (dict): Run parameters. Defaults to None.

        Returns:
            bool: True if it is downloaded.
        """

        params = params if params else {}
        params["format"] = format
        headers = Headers({"content-type": "application/json", "accept": "*/*"})
        data = self.connection.api_call(
            "GET",
            ["resources", self.id, "content"],
            params=params,
            stream=True,
            headers=headers,
        )

        with open(dest, "w") as local_file:
            for line in data.iter_lines():
                if line:
                    dcd_line = line.decode("utf-8")
                    local_file.write(dcd_line + "\n")
        data.close()
        return True



"""Module contains Identity model."""

from typing import Any, Dict  # noqa: F401

from crux.models.model import CruxModel


class Identity(CruxModel):
    """Identity Model."""

    def __init__(
        self,
        identity_id=None,  # type: str
        parent_identity_id=None,  # type: str
        description=None,  # type: str
        company_name=None,  # type: str
        first_name=None,  # type: str
        last_name=None,  # type: str
        role=None,  # type: str
        phone=None,  # type: str
        email=None,  # type: str
        type=None,  # type: str # type name is by design pylint: disable=redefined-builtin
        website=None,  # type: str
        landing_page=None,  # type: str
        connection=None,
        raw_response=None,  # type: Dict[Any, Any]
    ):
        # type: (...) -> None
        """
        Attributes:
            identity_id (str): Identity_id. Defaults to None.
            parent_identity_id (str): Parent Identity ID. Defaults to None.
            description (str): Description. Defaults to None.
            company_name (str): Company Name. Defaults to None.
            first_name (str): First Name. Defaults to None.
            last_name (str): Last Name. Defaults to None.
            role (str): Role. Defaults to None.
            phone (str): Phone. Defaults to None.
            email (str): Email. Defaults to None.
            type (str): Type. Defaults to None.
            website (str): Website. Defaults to None.
            landing_page (str): Landing page. Defaults to None.
            connection (crux._client.CruxClient): Connection Object. Defaults to None.
            raw_response (dict): Response Content. Defaults to None.

        Raises:
            ValueError: If identity_id is none.
        """
        self._identity_id = identity_id
        self._parent_identity_id = parent_identity_id
        self._description = description
        self._company_name = company_name
        self._first_name = first_name
        self._last_name = last_name
        self._role = role
        self._phone = phone
        self._email = email
        self._type = type
        self._website = website
        self._landing_page = landing_page

        self.connection = connection
        self.raw_response = raw_response

    @property
    def identity_id(self):
        """str: Gets the Identity Id."""
        return self._identity_id

    @property
    def parent_identity_id(self):
        """str: Gets the Parent Identity Id."""
        return self._parent_identity_id

    @property
    def company_name(self):
        """str: Gets the Company name."""
        return self._company_name

    @property
    def description(self):
        """str: Gets the Description."""
        return self._description

    @property
    def first_name(self):
        """str: Gets the First name."""
        return self._first_name

    @property
    def last_name(self):
        """str: Gets the Last name."""
        return self._last_name

    @property
    def role(self):
        """str: Gets the Role."""
        return self._role

    @property
    def email(self):
        """str: Gets the Email."""
        return self._email

    @property
    def website(self):
        """str: Gets the Website."""
        return self._website

    @property
    def landing_page(self):
        """str: Gets the Landing Page."""
        return self._landing_page

    @property
    def type(self):
        """str: Gets the Type."""
        return self._type

    @property
    def phone(self):
        """str: Gets the phone."""
        return self._phone

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Identity object to Identity Dictionary.

        Returns:
            dict: Identity Dictionary.
        """
        return {
            "identityId": self.identity_id,
            "parentIdentityId": self.parent_identity_id,
            "description": self.description,
            "companyName": self.company_name,
            "firstName": self.first_name,
            "lastName": self.last_name,
            "role": self.role,
            "phone": self.phone,
            "email": self.email,
            "website": self.website,
            "landingPage": self.landing_page,
            "type": self.type,
        }

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Identity
        """Transforms Identity Dictionary to Identity object.

        Args:
            a_dict (dict): Identity Dictionary.

        Returns:
            crux.models.Identity: Identity Object.
        """
        identity_id = a_dict["identityId"]
        # parent_identity_id = a_dict["parentIdentityId"]
        description = a_dict["description"]
        company_name = a_dict["companyName"]
        first_name = a_dict["firstName"]
        last_name = a_dict["lastName"]
        role = a_dict["role"]
        phone = a_dict["phone"]
        website = a_dict["website"]
        email = a_dict["email"]
        website = a_dict["website"]
        landing_page = a_dict["landingPage"]
        type = a_dict[  # type name is by design pylint: disable=redefined-builtin
            "type"
        ]

        return cls(
            identity_id=identity_id,
            description=description,
            company_name=company_name,
            first_name=first_name,
            last_name=last_name,
            role=role,
            phone=phone,
            email=email,
            website=website,
            landing_page=landing_page,
            type=type,
        )



"""Module defines abstract CruxModel."""

import pprint


class CruxModel(object):
    """Absract Crux Model."""

    def to_dict(self):
        """Absract to_dict method."""

    def to_str(self):
        # type: () -> str
        """Absract to_str method."""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        # type: () -> str
        """Absract __repr__ method."""
        return self.to_str()



"""Module contains Permission model."""

from typing import Dict  # noqa: F401

from crux.models.model import CruxModel


class Permission(CruxModel):
    """Permission Model."""

    def __init__(self, target_id=None, identity_id=None, permission_name=None):
        # type(str, str, str) -> None
        """
        Attributes:
            target_id (str): Target ID. Defaults to None.
            identity_id (str): Identity ID. Defaults to None.
            permission_name (str): Permission name. Defaults to None.
        """
        self._target_id = target_id
        self._identity_id = identity_id
        self._permission_name = permission_name

    @property
    def target_id(self):
        """str: Gets the Target ID."""
        return self._target_id

    @property
    def identity_id(self):
        """str: Gets the Identity ID."""
        return self._identity_id

    @property
    def permission_name(self):
        """str: Gets the Permission Name."""
        return self._permission_name

    def to_dict(self):
        # type: () -> Dict[str, str]
        """Transforms Dataset object to Dataset Dictionary.

        Returns:
            dict: Dataset Dictionary.
        """
        return {
            "targetId": self.target_id,
            "identityId": self.identity_id,
            "permissionName": self.permission_name,
        }

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Permission
        """Transforms Dataset Dictionary to Dataset object.

        Args:
            a_dict (dict): Dataset Dictionary.

        Returns:
            crux.models.Permission: Permission Object.
        """
        target_id = a_dict["targetId"]
        identity_id = a_dict["identityId"]
        permission_name = a_dict["permissionName"]

        return cls(
            target_id=target_id,
            identity_id=identity_id,
            permission_name=permission_name,
        )



"""Module contains Dataset model."""

import os
import posixpath
from typing import Any, Dict, IO, Iterator, List, Tuple, Union  # noqa: F401

from crux._compat import unicode
from crux._utils import create_logger, Headers, split_posixpath_filename_dirpath
from crux.exceptions import CruxAPIError, CruxClientError, CruxResourceNotFoundError
from crux.models._factory import get_resource_object
from crux.models.file import File
from crux.models.folder import Folder
from crux.models.job import LoadJob, StitchJob
from crux.models.label import Label
from crux.models.model import CruxModel
from crux.models.permission import Permission
from crux.models.query import Query
from crux.models.resource import Resource
from crux.models.table import Table


log = create_logger(__name__)


class Dataset(CruxModel):
    """Dataset Model."""

    def __init__(
        self,
        id=None,  # type: str # id name is by design pylint: disable=redefined-builtin
        owner_identity_id=None,  # type: str
        contact_identity_id=None,  # type: str
        name=None,  # type: str
        description=None,  # type: str
        website=None,  # type: str
        created_at=None,  # type: str
        modified_at=None,  # type: str
        connection=None,
        raw_response=None,  # type: Dict[Any, Any]
        tags=None,  # type: List[str]
    ):
        # type: (...) -> None
        """
        Attributes:
            id (str): Dataset Id. Defaults to None.
            owner_identity_id (str): Owner Identity Id. Defaults to None.
            contact_identity_id (str): Contact Identity Id. Defaults to None.
            name (str): Dataset name. Defaults to None.
            description (str): Dataset description. Defaults to None.
            website (str): Dataset website. Defaults to None.
            created_at (str): Dataset created. Defaults to None.
            modified_at (str): Dataset Modified. Defaults to None.
            connection (crux._client.CruxClient): Connection Object. Defaults to None.
            raw_response (dict): Response Content. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags to be applied to dataset.
                Defaults to None.

        Raises:
            ValueError: If name or tags are set to None.
        """
        self._id = id
        self._owner_identity_id = owner_identity_id
        self._contact_identity_id = contact_identity_id
        self._name = name
        self._description = description
        self._website = website
        self._created_at = created_at
        self._modified_at = modified_at
        self._tags = tags
        self._provenance = None

        self.connection = connection
        self.raw_response = raw_response

    @property
    def id(self):
        """str: Gets the Dataset ID."""
        return self._id

    @property
    def owner_identity_id(self):
        """str: Gets the Owner Identity ID."""
        return self._owner_identity_id

    @property
    def contact_identity_id(self):
        """str: Gets the Contact Identity ID."""
        return self._contact_identity_id

    @property
    def name(self):
        """str: Gets the Dataset Name."""
        return self._name

    @property
    def tags(self):
        """str: Gets the tags."""
        return self._tags

    @property
    def description(self):
        """str: Gets the Dataset Description."""
        return self._description

    @property
    def website(self):
        """str: Gets the Dataset Website."""
        return self._website

    @property
    def created_at(self):
        """str: Gets the Dataset created_at."""
        return self._created_at

    @property
    def modified_at(self):
        """str: Gets the Dataset modified_at."""
        return self._modified_at

    @property
    def provenance(self):
        """str: Compute or Get the provenance."""
        if self._provenance:
            return self._provenance

        self._provenance = self._get_provenance()
        return self._provenance

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Transforms Dataset object to Dataset Dictionary.

        Returns:
            dict: Dataset Dictionary.
        """
        return {
            "datasetId": self.id,
            "ownerIdentityId": self.owner_identity_id,
            "contactIdentityId": self.contact_identity_id,
            "description": self.description,
            "name": self.name,
            "website": self.website,
            "createdAt": self.created_at,
            "modifiedAt": self.modified_at,
            "tags": self.tags,
        }

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, Any]) -> Dataset
        """Transforms Dataset Dictionary to Dataset object.

        Args:
            a_dict (dict): Dataset Dictionary.

        Returns:
            crux.models.Dataset: Dataset Object.
        """
        id = a_dict[  # id name is by design pylint: disable=redefined-builtin
            "datasetId"
        ]
        owner_identity_id = a_dict["ownerIdentityId"]
        contact_identity_id = a_dict["contactIdentityId"]
        description = a_dict["description"]
        name = a_dict["name"]
        website = a_dict["website"]
        created_at = a_dict["createdAt"]
        modified_at = a_dict["modifiedAt"]
        tags = a_dict["tags"]

        return cls(
            id=id,
            owner_identity_id=owner_identity_id,
            contact_identity_id=contact_identity_id,
            description=description,
            name=name,
            website=website,
            created_at=created_at,
            modified_at=modified_at,
            tags=tags,
        )

    def _get_provenance(self):
        # type: () -> str
        """Fetches the provenance of the Dataset.

        Returns:
            list: List of provenance dictionaries containing DatasetID, WorkFlowID,
                systemOfRecord, cronSpec, createdAt and modifiedAt keys.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        response = self.connection.api_call(
            "GET", ["datasets", self.id, "provenance"], headers=headers
        )

        return response.json()

    def delete(self):
        # type: () -> bool
        """Deletes the dataset.

        Returns:
            bool: True if dataset is deleted.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "DELETE", ["datasets", self.id], headers=headers
        )

    def update(self, name=None, description=None, tags=None):
        # type: (str, str, List[str]) -> bool
        """Updates the metadata of dataset.

        Args:
            name (str): Name of the dataset. Defaults to None.
            description (str): Description of the dataset. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags. Defaults to None.

        Returns:
            bool: True, if dataset is updated.

        Raises:
            ValueError: It is raised if name, description or tags are unset.
            TypeError: It is raised if tags is not of type list.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        body = {}  # type: Dict[str, Union[str, List]]
        if name is not None:
            body["name"] = name
        if description is not None:
            body["description"] = description
        if tags is not None:
            if isinstance(tags, list):
                body["tags"] = tags
            else:
                raise TypeError("Tags should be of type list")

        if body:
            dataset_object = self.connection.api_call(
                "PUT", ["datasets", self.id], headers=headers, json=body, model=Dataset
            )
            self.__dict__.update(dataset_object.__dict__)
            log.debug(
                "Updated dataset %s with content %s", self.id, dataset_object.__dict__
            )
            return True
        else:
            raise ValueError("Name, Description or Tags should be set")

    def create_file(self, path, tags=None, description=None):
        # type: (str, List[str], str) -> File
        """Creates File resource in Dataset.

        Args:
            path (str): Path of the file resource.
            tags (:obj:`list` of :obj:`str`): Tags of the file resource.
                Defaults to None.
            description (str): Description of the file resource.
                Defaults to None.

        Returns:
            crux.models.File: File Object.
        """

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        tags = tags if tags else []

        file_name, folder = split_posixpath_filename_dirpath(path)

        file_resource = File(
            name=file_name,
            type="file",
            tags=tags,
            description=description,
            folder=folder,
        )

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=file_resource.to_dict(),
            model=File,
            headers=headers,
        )

    def create_table(self, path, config, tags=None, description=None):
        # type: (str, Dict[str, Any], List[str], str) -> Table
        """Creates Table resource in Dataset.

        Args:
            path (str): Table resource Path.
            config (dict): Table Schema Configuration.
            tags (:obj:`list` of :obj:`str`): Tags of the Table resource.
                Defaults to None.
            description (str): Description of the Table resource.
                Defaults to None.

        Returns:
            crux.models.Table: Table Object
        """

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        table_name, folder = split_posixpath_filename_dirpath(path)

        tags = tags if tags else []

        table_resource = Table(
            name=table_name,
            type="table",
            tags=tags,
            description=description,
            config=config,
            folder=folder,
        )

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=table_resource.to_dict(),
            model=Table,
            headers=headers,
        )

    def create_folder(self, path, folder="/", tags=None, description=None):
        # type: (str, str, List[str], str) -> Folder
        """Creates Folder resource in Dataset.

        Args:
            path (str): Path of the Folder resource.
            folder (str): Parent folder of the Folder resource.
                Defaults to /.
            tags (:obj:`list` of :obj:`str`): Tags of the Folder resource.
                Defaults to None.
            description (str): Description of the Folder resource.
                Defaults to None.

        Returns:
            crux.models.Folder: Folder Object.
        """

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        tags = tags if tags else []

        file_name, folder = split_posixpath_filename_dirpath(path)

        folder_resource = Folder(
            name=file_name,
            type="folder",
            tags=tags,
            description=description,
            folder=folder,
        )

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=folder_resource.to_dict(),
            model=Folder,
            headers=headers,
        )

    def _get_resource(self, path, model):
        """Gets the resource object from the string path.

        Args:
            path (str): Resource path.
            crux.models.Resource: Resource model.

        Returns:
            crux.models.Resource: Resource model Object.

        Raises:
            crux.exceptions.CruxResourceNotFoundError: If resource is not found.
        """
        resource_name, folder_path = split_posixpath_filename_dirpath(path)
        rsrc_list = self._list_resources(
            folder=folder_path,
            limit=1,
            name=resource_name,
            include_folders=True,
            model=model,
        )

        if rsrc_list:
            return rsrc_list[0]
        else:
            # As of now API can't fetch id from the resource path or name,
            # hence raising the 404 error from the Python client
            raise CruxResourceNotFoundError({"statusCode": 404, "name": resource_name})

    def _resource_exists(self, path):
        # type: (str) -> bool
        """Checks the existence of resource.

        Args:
            path (str): Resource Path.

        Returns:
            bool: True if resource exists.

        Raises:
            crux.exceptions.CruxResourceNotFoundError: If resource is not found.
        """
        try:
            self._get_resource(path=path, model=Resource)
            return True
        except CruxResourceNotFoundError:
            return False

    def get_file(self, path):
        # type: (str) -> File
        """Gets the File resource object.

        Args:
            path (str): File resource path.

        Returns:
            crux.models.File: File Object.
        """
        return self._get_resource(path=path, model=File)

    def get_folder(self, path):
        # type: (str) -> Folder
        """Gets the Folder resource object.

        Args:
            path (str): Folder resource path.

        Returns:
            crux.models.Folder: Folder Object.
        """
        return self._get_resource(path=path, model=Folder)

    def get_table(self, path):
        # type: (str) -> Table
        """Method which gets the Table resource

        Args:
            path: Table resource path

        Returns:
            crux.models.Table: Table Object
        """
        return self._get_resource(path=path, model=Table)

    def get_query(self, path):
        # type: (str) -> Query
        """Gets the Query resource object.

        Args:
            path (str): Query resource path.

        Returns:
            crux.models.Query: Query Object.
        """
        return self._get_resource(path=path, model=Query)

    def list_resources(
        self, folder="/", offset=0, limit=1, include_folders=False, sort=None
    ):
        # type: (str, int, int, bool, str) -> Resource
        """Lists the resources in Dataset.

        Args:
            folder (str): Folder for which resource should be listed.
                Defaults to /.
            offset (int): Sets the offset. Defaults to 0.
            limit (int): Sets the limit. Defaults to 1.
            include_folders (bool): Sets whether to include folders or not.
                Defaults to False.
            sort (str): Sets whether to sort or not.
                Defaults to None.

        Returns:
            list (:obj:`crux.models.Resource`): List of File resource objects.
        """
        return self._list_resources(
            sort=sort,
            folder=folder,
            offset=offset,
            limit=limit,
            include_folders=include_folders,
            model=Resource,
        )

    def download_files(self, folder, local_path, only_use_crux_domains=None):
        # type: (str, str, bool) -> List[str]
        """Downloads the resources recursively.

        Args:
            folder (str): Crux Dataset Folder from where the
                file resources should be recursively downloaded.
            local_path (str): Local OS Path where the file resources should be downloaded.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            list (:obj:`str`): List of location of download files.

        Raises:
            ValueError: If Folder or local_path is None.
            OSError: If local_path is an invalid directory location.
        """
        if folder is None:
            raise ValueError("Folder value shouldn't be empty")

        if local_path is None:
            raise ValueError("Local Path value shouldn't be empty")

        if not os.path.exists(local_path) and not os.path.isdir(local_path):
            raise OSError("local_path is an invalid directory location")

        local_file_list = []  # type: List[str]

        resources = self._list_resources(
            sort=None,
            folder=folder,
            offset=0,
            limit=None,
            include_folders=True,
            model=Resource,
        )

        for resource in resources:
            resource_path = posixpath.join(folder, resource.name)
            resource_local_path = os.path.join(local_path, resource.name)
            if resource.type == "folder":
                os.mkdir(resource_local_path)
                log.debug("Created local directory %s", resource_local_path)
                local_file_list += self.download_files(
                    folder=resource_path,
                    local_path=resource_local_path,
                    only_use_crux_domains=only_use_crux_domains,
                )
            elif resource.type == "file":
                file_resource = File.from_dict(resource.to_dict())
                file_resource.connection = self.connection
                file_resource.download(
                    resource_local_path, only_use_crux_domains=only_use_crux_domains
                )
                local_file_list.append(resource_local_path)
                log.debug("Downloaded file at %s", resource_local_path)

        return local_file_list

    def upload_files(
        self,
        local_path,
        folder,
        media_type=None,
        description=None,
        tags=None,
        only_use_crux_domains=None,
    ):
        # type: (str, str, str, str, List[str], bool) -> List[File]
        """Uploads the resources recursively.

        Args:
            local_path (str): Local OS Path from where the file resources should be uploaded.
            media_type (str): Content Types of File resources to be uploaded.
                Defaults to None.
            folder (str): Crux Dataset Folder where file resources
                should be recursively uploaded.
            description (str): Description to be set on uploaded resources.
                Defaults to None.
            tags (:obj:`list` of :obj:`str`): Tags to be set on uploaded resources.
                Defaults to None.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            list (:obj:`crux.models.File`): List of uploaded file objects.

        Raises:
            ValueError: If folder or local_path is None.
            OSError: If local_path is an invalid directory location.
        """
        tags = tags if tags else []

        uploaded_file_objects = []  # type: List[File]

        if folder is None:
            raise ValueError("Folder value shouldn't be empty")

        if local_path is None:
            raise ValueError("Local Path value shouldn't be empty")

        if not os.path.exists(local_path) and not os.path.isdir(local_path):
            raise OSError("local_path is an invalid directory location")

        for content in os.listdir(local_path):
            content_local_path = os.path.join(local_path, content)
            content_path = posixpath.join(folder, content)
            if os.path.isdir(content_local_path):
                self.create_folder(
                    path=content_path, tags=tags, description=description
                )
                log.debug("Created folder %s in dataset %s", content_path, self.id)
                uploaded_file_objects += self.upload_files(
                    media_type=media_type,
                    folder=content_path,
                    local_path=content_local_path,
                    tags=tags,
                    description=description,
                    only_use_crux_domains=only_use_crux_domains,
                )

            elif os.path.isfile(content_local_path):
                fil_o = self.upload_file(
                    content_local_path,
                    content_path,
                    media_type=media_type,
                    tags=tags,
                    description=description,
                    only_use_crux_domains=only_use_crux_domains,
                )
                uploaded_file_objects.append(fil_o)
                log.debug("Uploaded file %s in dataset %s", content_path, self.id)

        return uploaded_file_objects

    def list_files(self, sort=None, folder="/", offset=0, limit=100):
        # type: (str, str, int, int) -> List[File]
        """Lists the files.

        Args:
            sort (str): Sets whether to sort or not. Defaults to None.
            folder (str): Folder for which resource should be listed.
                Defaults to /.
            offset (int): Sets the offset. Defaults to 0.
            limit (int): Sets the limit. Defaults to 100.

        Returns:
            list (:obj:`crux.models.File`): List of File objects.
        """
        resource_list = self._list_resources(
            sort=sort,
            folder=folder,
            offset=offset,
            limit=limit,
            include_folders=False,
            model=File,
        )

        file_resource_list = []

        for resource in resource_list:
            if resource.type == "file":
                file_resource_list.append(resource)
        return file_resource_list

    def _list_resources(
        self,
        folder="/",
        offset=0,
        limit=1,
        include_folders=False,
        name=None,
        model=None,
        sort=None,
    ):

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        params = {"folder": folder, "offset": offset, "limit": limit}

        if sort:
            params["sort"] = sort

        if name:
            params["name"] = name

        if include_folders:
            params["includeFolders"] = "true"
        else:
            params["includeFolders"] = "false"

        return self.connection.api_call(
            "GET",
            ["datasets", self.id, "resources"],
            params=params,
            model=model,
            headers=headers,
        )

    def load_table_from_file(self, source_file, dest_table, append=False):
        # type: (str, str, bool) -> LoadJob
        """Loads table from file resource.

        Args:
            source_file (str or file): Source File Path in string or File Object.
            dest_table (str or crux.models.Table): Destination File Path in
                string or Table Object.
            append (bool): Sets whether to append to existing table. Defaults to False.

        Returns:
            crux.models.LoadJob: LoadJob Object.

        Raises:
            TypeError: If source_file or dest_table is not file or string object.
        """

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        if isinstance(source_file, File):
            src_file = source_file
        elif isinstance(source_file, str):
            src_file = self._get_resource(path=source_file, model=File)
        else:
            raise TypeError(
                "Invalid Type. It should be file path string or File resource object"
            )

        if isinstance(dest_table, Table):
            dst_table = dest_table
        elif isinstance(dest_table, (str, unicode)):
            dst_table = self._get_resource(path=dest_table, model=Table)
        else:
            raise TypeError(
                "Invalid Type. It should be path string or Table resource object"
            )

        payload = {
            "sourceId": src_file.id,
            "destinationId": dst_table.id,
            "append": append,
        }

        return self.connection.api_call(
            "POST",
            ["jobs", "loadtablefromfileresource"],
            json=payload,
            model=LoadJob,
            headers=headers,
        )

    def upload_file(
        self,
        src,
        dest,
        media_type=None,
        description=None,
        tags=None,
        only_use_crux_domains=None,
    ):
        # type: (Union[IO, str], str, str, str, List[str], bool) -> File
        """Uploads the File.

        Args:
            src (str or file): Local OS path whose content
                is to be uploaded to file resource.
            dest (str): File resource path.
            media_type (str): Content type of the file. Defaults to None.
            description (str): Description of the file. Defaults to None.
            tags (:obj:`list` of :obj:`str`): Tags to be attached to the file resource.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            crux.models.File: File Object.
        """
        tags = tags if tags else []

        file_resource = self.create_file(tags=tags, description=description, path=dest)

        try:
            return file_resource.upload(
                src, media_type=media_type, only_use_crux_domains=only_use_crux_domains
            )
        except (CruxClientError, CruxAPIError, IOError):
            file_resource.delete()
            raise

    def create_query(self, path, config, tags=None, description=None):
        # type: (str, Dict[str, Any], List[str], str) -> Query
        """Creates Query resource in Dataset.

        Args:
            path (str): Query resource Path.
            config (dict): Query configuration.
            tags (:obj:`list` of :obj:`str`): Tags of the Query resource.
                Defaults to None.
            description (str): Description of the Query resource.
                Defaults to None.

        Returns:
            crux.models.Query: Query Object.
        """

        query_name, folder = split_posixpath_filename_dirpath(path)

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        tags = tags if tags else []

        query_resource = Query(
            name=query_name,
            type="query",
            tags=tags,
            description=description,
            config=config,
            folder=folder,
        )

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=query_resource.to_dict(),
            model=Query,
            headers=headers,
        )

    def upload_query(self, sql_file, path, description=None, tags=None):
        # type: (str, str, str, List[str]) -> Query
        """Uploads the Query File.

        Args:
            path (str): Query resource path.
            sql_file (str): Local OS SQL file to be uploaded as query resource.
            description (str): Description for the Query resource.
                Defaults to None.
            tags (:obj:`list` of :obj:`str`): Tags for the Query resource.
                Defaults to None.

        Returns:
            crux.models.Query: Query Object.
        """

        with open(sql_file, mode="r") as data:
            query_config = {"query": data.read()}
            return self.create_query(
                path=path, tags=tags, description=description, config=query_config
            )

    def add_permission_to_resources(
        self,
        identity_id,
        permission,
        resource_paths=None,
        resource_objects=None,
        resource_ids=None,
    ):
        # type: (str, str, List[str], List[Union[File,Folder,Table,Query]], List[str]) -> bool
        """Adds permission to all or specific Dataset resources.

        Args:
            identity_id (str): Identity Id to be set.
            permission (str): Permission to be set.
            resource_paths (:obj:`list` of :obj:`str`): List of resource paths on which the
                permission should be applied. If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will apply the permission to whole dataset.
            resource_objects (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource objects on which the permission should be applied.
                Overrides resource_paths.
                If none of resource_paths, resource_objects or
                resource_ids parameter is set, then it will apply the
                permission to whole dataset.
            resource_ids (:obj:`list` of :obj:`str`): List of resource ids on which
                permission should be applied. Overrides resource_pathss and resource_objects.
                If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will apply the permission to whole dataset.

        Returns:
            bool: True if permission is applied.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "add",
        }  # type: Dict[str, Union[str, List[str]]]

        if resource_ids or resource_objects or resource_paths:
            if resource_paths:
                log.debug(
                    "Add permission %s to %s for resource paths",
                    permission,
                    identity_id,
                )
                resource_ids = list()
                for resource_path in resource_paths:
                    resource_object = self._get_resource(
                        path=resource_path, model=Resource
                    )
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_objects:
                log.debug(
                    "Add permissions %s to %s for resource objects",
                    permission,
                    identity_id,
                )
                resource_ids = list()
                for resource_object in resource_objects:
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_ids:
                log.debug(
                    "Add permissions %s to %s for resource ids", permission, identity_id
                )
                body["resourceIds"] = resource_ids
        else:
            log.debug(
                "Add permission %s to %s for dataset %s",
                permission,
                identity_id,
                self.id,
            )
            body["datasetId"] = self.id

        return self.connection.api_call(
            "POST", ["permissions", "bulk"], headers=headers, json=body
        )

    def delete_permission_from_resources(
        self,
        identity_id,
        permission,
        resource_paths=None,
        resource_objects=None,
        resource_ids=None,
    ):
        # type: (str, str, List[str], List[Union[File,Folder,Table,Query]], List[str]) -> bool
        """Method which deletes permission from all or specific Dataset resources.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.
            resource_paths (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource path from which the permission should be deleted.
                If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will delete the permission from whole dataset.
            resource_objects (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource objects from which the permission should be deleted.
                Overrides resource_paths.
                If none of resource_paths, resource_objects or resource_ids parameter is
                set, then it will delete the permission from whole dataset.
            resource_ids (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource ids from which the permission should be deleted.
                Overrides resource_paths and resource_objects.
                If none of resource_paths, resource_objects or
                resource_ids parameter is set, then it will delete the
                permission from whole dataset.

        Returns:
            bool: True if it is able to delete the permission.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "delete",
        }  # type: Dict[str, Union[str, List[str]]]

        if resource_ids or resource_objects or resource_paths:
            if resource_paths:
                log.debug(
                    "Delete permission %s for %s from resource paths",
                    permission,
                    identity_id,
                )
                resource_ids = list()
                for resource_path in resource_paths:
                    resource_object = self._get_resource(
                        path=resource_path, model=Resource
                    )
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_objects:
                log.debug(
                    "Delete permission %s for %s from resource objects",
                    permission,
                    identity_id,
                )
                resource_ids = list()
                for resource_object in resource_objects:
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_ids:
                log.debug(
                    "Delete permission %s for %s from resource ids",
                    permission,
                    identity_id,
                )
                body["resourceIds"] = resource_ids
        else:
            log.debug(
                "Delete permission %s for %s from dataset %s",
                permission,
                identity_id,
                self.id,
            )
            body["datasetId"] = self.id

        return self.connection.api_call(
            "POST", ["permissions", "bulk"], headers=headers, json=body
        )

    def add_permission(self, identity_id, permission):
        # type: (str, str) -> Union[bool, Permission]
        """Adds permission to the Dataset.

        Args:
            identity_id: Identity Id to be set.
            permission: Permission to be set.

        Returns:
            crux.models.Permission: Permission Object.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "PUT",
            ["permissions", self.id, identity_id, permission],
            model=Permission,
            headers=headers,
        )

    def delete_permission(self, identity_id, permission):
        # type: (str, str) -> bool
        """Deletes permission from the Dataset.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        return self.connection.api_call(
            "DELETE", ["permissions", self.id, identity_id, permission], headers=headers
        )

    def list_permissions(self):
        # type: () -> List[Permission]
        """Lists the permission on the Dataset.

        Returns:
            list (:obj:`crux.models.Permission`): List of Permission Objects.
        """
        headers = {"Accept": "application/json"}
        return self.connection.api_call(
            "GET",
            ["datasets", self.id, "permissions"],
            model=Permission,
            headers=headers,
        )

    def add_label(self, label_key, label_value):
        # type: (str, str) -> bool
        """Adds label to Dataset.

        Args:
            label_key (str): Label Key for Dataset.
            label_value (str): Label Value for Dataset.

        Returns:
            bool: True if labels are added.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "PUT",
            ["datasets", self.id, "labels", label_key, label_value],
            headers=headers,
        )

    def delete_label(self, label_key):
        # type: (str) -> bool
        """Deletes label from Dataset.

        Args:
            label_key (str): Label Key for Dataset.

        Returns:
            bool: True if labels are deleted.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "DELETE", ["datasets", self.id, "labels", label_key], headers=headers
        )

    def get_label(self, label_key):
        # type: (str) -> Label
        """Gets label value of Dataset.

        Args:
            label_key (str): Label Key for Dataset.

        Returns:
            crux.models.Label: Label Object.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "GET",
            ["datasets", self.id, "labels", label_key],
            headers=headers,
            model=Label,
        )

    def find_resources_by_label(self, predicates, max_per_page=1000):
        # type: (List[Dict[str,str]],int)->Iterator[Union[File,Folder,Query,Table]]
        """Method which searches the resouces for given labels in Dataset

        Each predicate can be either:

        - Lexicographical equal
        - Lexicographical less than
        - Lexicographical less than or equal to
        - Lexicographical greater than
        - Lexicographical greater than or equal to
        - A list of OR predicates
        - A list of AND predicates

        .. code-block:: python

            predicates = [
                {"op": "eq", "key": "key1", "val": "abcd"},
                {"op": "ne", "key": "key1", "val": "zzzz"},
                {"op": "lt", "key": "key1", "val": "abd"},
                {"op": "gt", "key": "key1", "val": "abc"},
                {"op": "lte", "key": "key1", "val": "abd"},
                {"op": "gte", "key": "key1", "val": "abc"},
                {"op": "or", "in":
                    [
                        {"op": "eq", "key": "key1", "val": "abcd"},
                        # more OR predicates...
                    ]
                },
                {"op": "and", "in":
                    [
                        {"op": "eq", "key": "key1", "val": "abcd"},
                        # more AND predicates...
                    ]
                }
            ]

        Args:
            predicates (:obj:`list` of :obj:`dict`): List of dictionary predicates
                for finding resources.
            max_per_page (int): Pagination limit. Defaults to 1000.

        Returns:
            list (:obj:`crux.models.Resource`): List of resource matching the query parameters.

        Example:
            .. code-block:: python

                from crux import Crux

                conn = Crux()
                dataset_object = conn.get_dataset(id="dataset_id")
                predicates=[
                    {"op":"eq","key":"test_label1","val":"test_value1"}
                ]
                resource_objects = dataset_object.find_resources_by_label(
                    predicates=predicates
                )
        """

        predicates = predicates if predicates else []

        query_params = {"limit": max_per_page}

        predicates_query = {
            "basic_query": predicates
        }  # type: Dict[str, List[Dict[str,str]]]

        after = None
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        while True:

            if after:
                query_params["after"] = after

            response = self.connection.api_call(
                "POST",
                ["datasets", self.id, "labels", "search"],
                headers=headers,
                json=predicates_query,
                params=query_params,
            )

            resource_list = response.json().get("results")
            if resource_list:
                after = resource_list[-1].get("resourceId")
                for resource in resource_list:
                    obj = get_resource_object(
                        resource_type=resource.get("type"), data=resource
                    )
                    obj.connection = self.connection
                    obj.raw_response = resource
                    yield obj
            else:
                return

    def stitch(
        self,
        source_resources,
        destination_resource,
        labels=None,
        tags=None,
        description=None,
    ):
        # type: (List[Union[str, File]], str, str, List[str], str) -> Tuple[File, str]
        """Method which stitches multiple Avro resources into single Avro resource

        Args:
            source_resources (:obj:`list` of :obj:`str` or :obj:`file`): List of resource paths
                which are to be stitched.
            destination_resource (str): Resource Path to load the stitched output
            labels (dict): Key/Value labels that should be applied to stitched resource
            tags (:obj:`list` of :obj:`str`): List of tags to be applied
                on destination resource.
                Taken into consideration if resource is required to be created.
            description (str): Description to be applied created destination.
                Taken into consideration if resource is required to be created.

        Returns:
            tuple (:obj:`crux.models.File`, :obj:`str`): File object of destination resource.
                    Job ID for background running job.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        source_resource_ids = list()
        for resource in source_resources:
            if isinstance(resource, File):
                log.debug("Stitch source resources are of type crux.models.File")
                source_resource_ids.append(resource.id)
            elif isinstance(resource, (str, unicode)):
                log.debug("Stitch source resources are of type string")
                resource_object = self._get_resource(path=resource, model=File)
                source_resource_ids.append(resource_object.id)
            else:
                raise TypeError(
                    "Invalid Type. It should be File resource object or path string"
                )

        if isinstance(destination_resource, File):
            log.debug("Stitch destination resource is of type crux.models.File")
            destination_file_object = destination_resource
        elif isinstance(destination_resource, str):
            log.debug("Stitch destination resource is of type string")
            if self._resource_exists(path=destination_resource):
                destination_file_object = self._get_resource(
                    path=destination_resource, model=File
                )
            else:
                log.debug("Creating file resource at %s", destination_resource)
                destination_file_object = self.create_file(
                    path=destination_resource,
                    description=description,
                    tags=tags if tags else [],
                )
        else:
            raise TypeError(
                "Invalid Type. It should be File resource object or path string"
            )

        data = {
            "sourceResourceIds": source_resource_ids,
            "destinationResourceId": destination_file_object.id,
            "labelsToApply": labels if labels else {},
        }
        response = self.connection.api_call(
            "POST", ["datasets", self.id, "stitch"], headers=headers, json=data
        )

        raw_json = response.json().get("destinationResource")

        file_object = Resource.from_dict(raw_json)

        file_object.connection = self.connection

        file_object.raw_response = raw_json

        job_id = response.json().get("jobId")

        return (file_object, job_id)

    def get_stitch_job(self, job_id):
        # type: (str) -> StitchJob
        """Stitch Job Details.

        Args:
            job_id (str): Job ID of the Stitch Job.

        Returns:
            crux.models.StitchJob: StitchJob object.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        return self.connection.api_call(
            "GET", ["datasets", "stitch", job_id], headers=headers, model=StitchJob
        )



"""
Setup file for Python packaging.
"""

import io
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
requirements = [
    "enum34;python_version<'3.4'",
    "google-resumable-media[requests]",
    "typing;python_version<'3.5'",
]
packages = [pkg for pkg in find_packages() if pkg.startswith("crux")]

version = {}
with io.open(os.path.join(here, "crux", "__version__.py"), "r", encoding="utf-8") as fh:
    exec(fh.read(), version)

with io.open("README.md", "r", encoding="utf-8") as fh:
    readme = fh.read()

setup(
    name="crux",
    packages=packages,
    version=version["__version__"],
    description="Crux Informatics API client library",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Crux Informatics, Inc.",
    author_email="python-opensource@cruxinformatics.com",
    url="https://github.com/cruxinformatics/crux-python",
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    license="MIT",
    install_requires=requirements,
    keywords=["crux-python"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development",
    ],
)



"""
Module containing models that represent objects returned by the API.
"""

import logging
from logging import NullHandler

from crux.models.dataset import Dataset
from crux.models.file import File
from crux.models.folder import Folder
from crux.models.identity import Identity
from crux.models.job import Job, LoadJob, StitchJob
from crux.models.label import Label
from crux.models.permission import Permission
from crux.models.query import Query
from crux.models.resource import Resource
from crux.models.table import Table


__all__ = (
    "Identity",
    "Permission",
    "LoadJob",
    "StitchJob",
    "Job",
    "Resource",
    "File",
    "Folder",
    "Table",
    "Dataset",
    "Query",
    "Label",
)

# Set default logging handler to avoid "No handler found" warnings
logging.getLogger(__name__).addHandler(NullHandler())



"""Module contains AbstractJob, Job, LoadJob Model."""

from typing import Dict  # noqa: F401

from crux.models.model import CruxModel


class AbstractJob(CruxModel):
    """AbstractJob Model."""


class LoadJob(AbstractJob):
    """LoadJob Model."""

    def __init__(self, job_id=None, job_url=None):
        # type(str, str) -> None
        """
        Attributes:
            job_id (str): Job ID. Defaults to None.
            job_url (str): Job Url. Defaults to None.
        """

        self._job_id = job_id
        self._job_url = job_url

    @property
    def job_id(self):
        """str: Gets the Job Id."""
        return self._job_id

    @property
    def job_url(self):
        """str: Gets the Job URL."""
        return self._job_url

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> LoadJob
        """Transforms LoadJob Dictionary to LoadJob object.

        Args:
            a_dict (dict): LoadJob Dictionary.

        Returns:
            crux.models.LoadJob: LoadJob Object.
        """
        job_id = a_dict["jobId"]
        job_url = a_dict["jobUrl"]

        return cls(job_id=job_id, job_url=job_url)


class Job(AbstractJob):
    """Job Model."""

    def __init__(self, job_id=None, status=None, statistics=None, connection=None):
        # type: (str, Status, Statistics, str) -> None
        """
        Attributes:
            job_id (str): Job ID. Defaults to None.
            status (Status): Status Object. Defaults to None.
            statistics (Statistics): Statistics Object. Defaults to None.
        """
        self.job_id = job_id
        self.status = status
        self.statistics = statistics
        self.connection = connection

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Job
        """Transforms Job Dictionary to Job object.

        Args:
            a_dict (dict): Job Dictionary.

        Returns:
            crux.models.Job: Job Object.
        """
        job_id = a_dict["jobId"]
        status = Status.from_dict(a_dict["status"])  # type: ignore
        statistics = Statistics.from_dict(a_dict["statistics"])  # type: ignore

        return cls(job_id=job_id, status=status, statistics=statistics)


class Status(object):
    """Job Status Model."""

    def __init__(self, state=None):
        # type: (str) -> None
        """
        Attributes:
            state (str): State. Defaults to None.
        """
        self.state = state

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Status
        """Transforms Job Status Dictionary to Job Status object.

        Args:
            a_dict (dict): Job Status Dictionary.

        Returns:
            crux.models.job.Status: Job Status Object.
        """
        state = a_dict["state"]

        return cls(state=state)


class Statistics(object):
    """Job Statistic Model."""

    def __init__(self, creation_time=None, start_time=None, end_time=None, load=None):
        # type(str, str, str, Load) -> None
        """
        Attributes:
            creation_time (str): Creation Time. Defaults to None.
            start_time (str): Start Time. Defaults to None.
            end_time (str): End Time. Defaults to None.
            load (Load): Load Object. Defaults to None.
        """
        self.creation_time = creation_time
        self.start_time = start_time
        self.end_time = end_time
        self.load = load

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Statistics
        """Transforms Job Statistics Dictionary to Job Statistics object.

        Args:
            a_dict (dict): Job Statistics Dictionary.

        Returns:
            crux.models.job.Statistics: Job Statistics Object.
        """
        creation_time = a_dict["creationTime"]
        start_time = a_dict["startTime"]
        end_time = a_dict["endTime"]
        load = Load.from_dict(a_dict["load"])  # type: ignore

        return cls(
            creation_time=creation_time,
            start_time=start_time,
            end_time=end_time,
            load=load,
        )


class Load(object):
    """ Job Load Model """

    def __init__(
        self,
        input_files=None,  # type: str
        input_file_bytes=None,  # type: str
        output_rows=None,  # type: str
        output_bytes=None,  # type: str
        bad_records=None,  # type: str
    ):
        # type(...) -> None
        """
        Attributes:
            input_files (str): Input files. Defaults to None.
            input_file_bytes (str): Input files bytes. Defaults to None.
            output_rows (str): Output rows. Defaults to None.
            output_bytes (str): Output Bytes. Defaults to None.
            bad_records (str): Bad Records. Defaults to None.
        """
        self.input_files = input_files
        self.input_file_bytes = input_file_bytes
        self.output_rows = output_rows
        self.output_bytes = output_bytes
        self.bad_records = bad_records

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> Load
        """Transforms Job Load Dictionary to Job Load object.

        Args:
            a_dict (dict): Job Load Dictionary.

        Returns:
            crux.models.job.Load: Job Load Object.
        """
        input_files = a_dict["inputFiles"]
        input_file_bytes = a_dict["inputFileBytes"]
        output_rows = a_dict["outputRows"]
        output_bytes = a_dict["outputBytes"]
        bad_records = a_dict["badRecords"]

        return cls(
            input_files=input_files,
            input_file_bytes=input_file_bytes,
            output_rows=output_rows,
            bad_records=bad_records,
            output_bytes=output_bytes,
        )


class StitchJob(AbstractJob):
    """Stitch Job Model."""

    def __init__(self, job_id=None, status=None):
        # type(str, str) -> None
        """
        Attributes:
            job_id (str): Job ID. Defaults to None.
            status (str): Status. Defaults to None.
        """
        self.job_id = job_id
        self.status = status

    @classmethod
    def from_dict(cls, a_dict):
        # type: (Dict[str, str]) -> StitchJob
        """Transforms Stitch Job Dictionary to Stitch Job object.

        Args:
            a_dict (dict): Stitch Job Dictionary.

        Returns:
            crux.models.job.StitchJob: Stitch Job Object.
        """
        job_id = a_dict["jobId"]
        status = a_dict["status"]

        return cls(job_id=job_id, status=status)



"""
Module packages root level crux objects.
"""
import logging
from logging import NullHandler

from crux._utils import TRACE
from crux.apis import Crux

__all__ = ("Crux", "TRACE")

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(NullHandler())
